import torch 
import torch.nn as nn
from medlat.registry import register_model
from medlat.first_stage.discrete.modules.ldm_modules import Encoder, Decoder
from einops.layers.torch import Rearrange, Reduce
from medlat.first_stage.discrete.vq_models import VQModel, VQModelTransformer
from medlat.first_stage.discrete.quantizer import *
from medlat.modules.alignments import *
from medlat.first_stage.token.maetok.modules.vit_models import MAETokViTEncoder, MAETokViTDecoder
from medlat.modules.vit_core import GenericViTEncoder, GenericViTDecoder

@register_model("discrete.vq.f4_d3_e8192")
def VQ_f4_d3_e8192(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=3,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        n_e=8192,
        e_dim=3,
        beta=0.25,
        use_ema=False,
        ema_decay=0.99,
        ema_eps=1e-5,
        **kwargs
    ):
    """
    Instantiate a VQModel (f4 config) with flexible, efficient parameter override.
    https://github.com/CompVis/latent-diffusion/blob/main/models/first_stage_models/vq-f4/config.yaml
    """
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = VectorQuantizer2(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        use_ema=use_ema,
        ema_decay=ema_decay,
        ema_eps=ema_eps
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.vq.f4_d8_e8192")
def VQ_f4_d8_e8192(**kwargs):
    return VQ_f4_d3_e8192(z_channels=8, e_dim=8, **kwargs)

@register_model(f"discrete.vq.f4_d16_e8192")
def VQ_f4_d16_e8192(**kwargs):
    return VQ_f4_d3_e8192(z_channels=16, e_dim=16, **kwargs)

@register_model(f"discrete.vq.f4_d32_e8192")
def VQ_f4_d32_e8192(**kwargs):
    return VQ_f4_d3_e8192(z_channels=32, e_dim=32, **kwargs)

@register_model(f"discrete.vq.f8_d4_e16384")
def VQ_f8_d4_e16384(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=4,
        beta=0.25,
        use_ema=False,
        ema_decay=0.99,
        ema_eps=1e-5,
        **kwargs
    ):
    """
    Instantiate a VQModel (f8 config) with flexible, efficient parameter override.
    https://github.com/CompVis/latent-diffusion/blob/main/models/first_stage_models/vq-f8/config.yaml
    """
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = VectorQuantizer2(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        use_ema=use_ema,
        ema_decay=ema_decay,
        ema_eps=ema_eps
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.vq.f8_d8_e16384")
def VQ_f8_d8_e16384(**kwargs):
    return VQ_f8_d4_e16384(z_channels=8, e_dim=8, **kwargs)

@register_model(f"discrete.vq.f8_d16_e16384")
def VQ_f8_d16_e16384(**kwargs):
    return VQ_f8_d4_e16384(z_channels=16, e_dim=16, **kwargs)

@register_model(f"discrete.vq.f8_d32_e16384")
def VQ_f8_d32_e16384(**kwargs):
    return VQ_f8_d4_e16384(z_channels=32, e_dim=32, **kwargs)

@register_model(f"discrete.vq.f16_d8_e16384")
def VQ_f16_d8_e16384(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=8,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[16],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=8,
        beta=0.25,
        use_ema=False,
        ema_decay=0.99,
        ema_eps=1e-5,
        **kwargs
    ):
    """
    Instantiate a VQModel (f16 config) with flexible, efficient parameter override.
    https://github.com/CompVis/latent-diffusion/blob/main/models/first_stage_models/vq-f16/config.yaml
    """
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = VectorQuantizer2(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        use_ema=use_ema,
        ema_decay=ema_decay,
        ema_eps=ema_eps
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.vq.f16_d16_e16384")
def VQ_f16_d16_e16384(**kwargs):
    return VQ_f16_d8_e16384(z_channels=16, **kwargs)

@register_model(f"discrete.vq.f16_d32_e16384")
def VQ_f16_d32_e16384(**kwargs):
    return VQ_f16_d8_e16384(z_channels=32, **kwargs)

@register_model(f"discrete.vq.f16_d64_e16384")
def VQ_f16_d64_e16384(**kwargs):
    return VQ_f16_d8_e16384(z_channels=64, **kwargs)

@register_model(f"discrete.maskgit.vq.f16_d256_e1024",
code_url="https://github.com/google-research/maskgit/blob/main/maskgit/nets/vqgan_tokenizer.py",
paper_url="https://arxiv.org/abs/2202.04200",)
def MaskGITVQ_f16_d256_e1024(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=256,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        n_e=1024,
        e_dim=256,
        beta=0.25,
        entropy_loss_ratio=0.1,
        entropy_loss_type="softmax",
        entropy_temperature=0.01,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = VectorQuantizer2(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        entropy_loss_ratio=entropy_loss_ratio,
        entropy_loss_type=entropy_loss_type,
        entropy_temperature=entropy_temperature,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.msrq.f16_d32_e4096",
code_url="https://github.com/FoundationVision/VAR/blob/main/models/quant.py",
paper_url="https://arxiv.org/pdf/2404.02905",)
def MSRQ_f16_d32_e4096(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=32,
        in_channels=3,
        out_ch=3,
        ch=160,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=4096,
        e_dim=32,
        beta=0.25,
        use_ema=False,
        ema_decay=0.99,
        ema_eps=1e-5,
        v_patch_nums=(1, 2, 3, 4, 5, 6, 8, 10, 13, 16),
        quant_resi=0.5,
        share_quant_resi=4,
        using_znorm=False,
        **kwargs
    ):
    """
    Instantiate a MSRQModel (EMA, f8 config) with flexible, efficient parameter override.
    """
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = MultiScaleResidualQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        v_patch_nums=v_patch_nums,
        quant_resi=quant_resi,
        share_quant_resi=share_quant_resi,
        using_znorm=using_znorm,
        dims=dims,
        use_ema=use_ema,
        ema_decay=ema_decay,
        ema_eps=ema_eps
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.lfq.f16_d10_b10")
def LFQ_f16_d10_b10(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=10,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=10,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = LookupFreeQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.lfq.f8_d10_b10")
def LFQ_f8_d10_b10(**kwargs):
    return LFQ_f16_d10_b10(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f4_d10_b10")
def LFQ_f4_d10_b10(**kwargs):
    return LFQ_f16_d10_b10(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f16_d12_b12")
def LFQ_f16_d12_b12(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=12,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=12,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = LookupFreeQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.lfq.f8_d12   _b12")
def LFQ_f8_d12_b12(**kwargs):
    return LFQ_f16_d12_b12(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f4_d12_b12")
def LFQ_f4_d12_b12(**kwargs):
    return LFQ_f16_d12_b12(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f16_d14_b14")
def LFQ_f16_d14_b14(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=14,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=14,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = LookupFreeQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.lfq.f8_d14_b14")
def LFQ_f8_d14_b14(**kwargs):
    return LFQ_f16_d14_b14(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f4_d14_b14")
def LFQ_f4_d14_b14(**kwargs):
    return LFQ_f16_d14_b14(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f16_d16_b16")
def LFQ_f16_d16_b16(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=16,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=16,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = LookupFreeQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.lfq.f8_d16_b16")
def LFQ_f8_d16_b16(**kwargs):
    return LFQ_f16_d16_b16(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f4_d16_b16")
def LFQ_f4_d16_b16(**kwargs):
    return LFQ_f16_d16_b16(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f16_d18_b18")
def LFQ_f16_d18_b18(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=18,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=18,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = LookupFreeQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.lfq.f8_d18_b18")
def LFQ_f8_d18_b18(**kwargs):
    return LFQ_f16_d18_b18(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.lfq.f4_d18_b18")
def LFQ_f4_d18_b18(**kwargs):
    return LFQ_f16_d18_b18(ch_mult=[1, 2, 4], **kwargs)

@register_model("discrete.simple_qinco.f4_d3_e8192")
def SimpleQINCo_f4_d3_e8192(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=3,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        n_e=8192,
        e_dim=3,
        beta=0.25,
        hidden_dim=256,
        num_layers=3,
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SimpleQINCo(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        hidden_dim=hidden_dim,
        num_layers=num_layers
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.simple_qinco.f8_d4_e16384")
def SimpleQINCo_f8_d4_e16384(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=4,
        beta=0.25,
        hidden_dim=256,
        num_layers=3,
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SimpleQINCo(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        hidden_dim=hidden_dim,
        num_layers=num_layers
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.simple_qinco.f16_d8_e16384")
def SimpleQINCo_f16_d8_e16384(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=8,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[16],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=8,
        beta=0.25,
        hidden_dim=256,
        num_layers=3,
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SimpleQINCo(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        hidden_dim=hidden_dim,
        num_layers=num_layers
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model("discrete.simvq.f4_d3_e8192")
def SimVQ_f4_d3_e8192(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=3,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        n_e=8192,
        e_dim=3,
        beta=0.25,
        commitment_weight=1.0,
        rotation_trick=True,
        codebook_transform=None,
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SimVQ(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        rotation_trick=rotation_trick,
        commitment_weight=commitment_weight,
        codebook_transform=codebook_transform,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.simvq.f8_d4_e16384")
def SimVQ_f8_d4_e16384(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=4,
        beta=0.25,
        commitment_weight=1.0,
        rotation_trick=True,
        codebook_transform=None,
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SimVQ(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        rotation_trick=rotation_trick,
        commitment_weight=commitment_weight,
        codebook_transform=codebook_transform,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.simvq.f16_d8_e16384")
def SimVQ_f16_d8_e16384(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=8,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[16],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=8,
        beta=0.25,
        commitment_weight=1.0,
        rotation_trick=True,
        codebook_transform=None,
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SimVQ(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        rotation_trick=rotation_trick,
        commitment_weight=commitment_weight,
        codebook_transform=codebook_transform,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model("discrete.fsq.f4_d3_l8192")
def FSQ_f4_d3_l8192(
        # --- encoder/decoder config (IDENTICAL) ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=3,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,

        # --- FSQ quantizer config ---
        levels=[32, 16, 16],  # 8192 total states
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )

    quantizer = FiniteScalarQuantizer(
        levels=levels,
        dim=z_channels,
    )

    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model("discrete.fsq.f8_d4_l16384")
def FSQ_f8_d4_l16384(
        # --- encoder/decoder config (IDENTICAL) ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,

        # --- FSQ quantizer config ---
        levels=[16, 8, 8, 16],  # 16384 total states
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )

    quantizer = FiniteScalarQuantizer(
        levels=levels,
        dim=z_channels,
    )

    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model("discrete.fsq.f16_d8_l16384")
def FSQ_f16_d8_l16384(
        # --- encoder/decoder config (IDENTICAL) ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=8,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[16],
        dropout=0.0,

        # --- FSQ quantizer config ---
        levels=[4, 4, 4, 4, 4, 4, 2, 2],  # 16384 total states
        **kwargs
    ):

    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )

    quantizer = FiniteScalarQuantizer(
        levels=levels,
        dim=z_channels,
    )

    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.fsq.f16_d256_e8",
code_url="https://github.com/google-research/google-research/blob/master/fsq/fsq.ipynb",
paper_url="https://arxiv.org/pdf/2309.15505",
description="The original version employs on maskgits VQVAE config")
def FSQ_f16_d256_e8(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=256,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        levels=[8, 6, 5],
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = FiniteScalarQuantizer(
        levels=levels,
        dim=z_channels,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.fsq.f16_d256_e10",
code_url="https://github.com/google-research/google-research/blob/master/fsq/fsq.ipynb",
paper_url="https://arxiv.org/pdf/2309.15505",
description="The original version employs on maskgits VQVAE config levels from table1")
def FSQ_f16_d256_e10(**kwargs):
    return FSQ_f16_d256_e8(levels=[8, 5, 5, 5], **kwargs)

@register_model(f"discrete.fsq.f16_d256_e12",
code_url="https://github.com/google-research/google-research/blob/master/fsq/fsq.ipynb",
paper_url="https://arxiv.org/pdf/2309.15505",
description="The original version employs on maskgits VQVAE config levels from table1")
def FSQ_f16_d256_e12(**kwargs):
    return FSQ_f16_d256_e8(levels=[7, 5, 5, 5, 5], **kwargs)

@register_model(f"discrete.fsq.f16_d256_e14",
code_url="https://github.com/google-research/google-research/blob/master/fsq/fsq.ipynb",
paper_url="https://arxiv.org/pdf/2309.15505",
description="The original version employs on maskgits VQVAE config levels from table1")
def FSQ_f16_d256_e14(**kwargs):
    return FSQ_f16_d256_e8(levels=[8, 8, 8, 6, 5], **kwargs)

@register_model(f"discrete.fsq.f16_d256_e16",
code_url="https://github.com/google-research/google-research/blob/master/fsq/fsq.ipynb",
paper_url="https://arxiv.org/pdf/2309.15505",
description="The original version employs on maskgits VQVAE config levels from table1")
def FSQ_f16_d256_e16(**kwargs):
    return FSQ_f16_d256_e8(levels=[8, 8, 8, 5, 5, 5], **kwargs)


@register_model(f"discrete.bsq.f16_d10_b10")
def BSQ_f16_d10_b10(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=10,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=10,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = BinarySphericalQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.bsq.f8_d10_b10")
def BSQ_f8_d10_b10(**kwargs):
    return BSQ_f16_d10_b10(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f4_d10_b10")
def BSQ_f4_d10_b10(**kwargs):
    return BSQ_f16_d10_b10(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f16_d12_b12")
def BSQ_f16_d12_b12(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=12,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=12,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = BinarySphericalQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.bsq.f8_d12_b12")
def BSQ_f8_d12_b12(**kwargs):
    return BSQ_f16_d12_b12(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f4_d12_b12")
def BSQ_f4_d12_b12(**kwargs):
    return BSQ_f16_d12_b12(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f16_d14_b14")
def BSQ_f16_d14_b14(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=14,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=14,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = BinarySphericalQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.bsq.f8_d14_b14")
def BSQ_f8_d14_b14(**kwargs):
    return BSQ_f16_d14_b14(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f4_d14_b14")
def BSQ_f4_d14_b14(**kwargs):
    return BSQ_f16_d14_b14(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f16_d16_b16")
def BSQ_f16_d16_b16(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=16,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=16,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = BinarySphericalQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.bsq.f8_d16_b16")
def BSQ_f8_d16_b16(**kwargs):
    return BSQ_f16_d16_b16(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f4_d16_b16")
def BSQ_f4_d16_b16(**kwargs):
    return BSQ_f16_d16_b16(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f16_d18_b18")
def BSQ_f16_d18_b18(
        # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=18,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        token_bits=18,
        commitment_cost=0.25,
        entropy_loss_weight=0.02,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = BinarySphericalQuantizer(
        token_bits=token_bits,
        commitment_cost=commitment_cost,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.bsq.f8_d18_b18")
def BSQ_f8_d18_b18(**kwargs):
    return BSQ_f16_d18_b18(ch_mult=[1, 2, 2, 4], **kwargs)

@register_model(f"discrete.bsq.f4_d18_b18")
def BSQ_f4_d18_b18(**kwargs):
    return BSQ_f16_d18_b18(ch_mult=[1, 2, 4], **kwargs)

@register_model(f"discrete.rqvae.f4_d3_e8192")
def RQVAE_f4_d3_e8192(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=3,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=VectorQuantizer2,
        num_quantizers=4,
        n_e=8192,
        e_dim=3,
        beta=0.25,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = ResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.rqvae.f8_d4_e16384")
def RQVAE_f8_d4_e16384(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=VectorQuantizer2,
        num_quantizers=4,
        n_e=16384,
        e_dim=4,
        beta=0.25,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = ResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.rqvae.f16_d8_e16384")
def RQVAE_f16_d8_e16384(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=8,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[16],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=VectorQuantizer2,
        num_quantizers=4,
        n_e=16384,
        e_dim=8,
        beta=0.25,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = ResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.rsimple_qinco.f4_d3_e8192")
def RSimpleQINCo_f4_d3_e8192(**kwargs):
    return RQVAE_f4_d3_e8192(**kwargs, quantizer_class=SimpleQINCo)

@register_model(f"discrete.rsimple_qinco.f8_d4_e16384")
def RSimpleQINCo_f8_d4_e16384(**kwargs):
    return RQVAE_f8_d4_e16384(**kwargs, quantizer_class=SimpleQINCo)

@register_model(f"discrete.rsimple_qinco.f16_d8_e16384")
def RSimpleQINCo_f16_d8_e16384(**kwargs):
    return RQVAE_f16_d8_e16384(**kwargs, quantizer_class=SimpleQINCo)


@register_model(f"discrete.qinco.f4_d3_e8192")
def QINCo_f4_d3_e8192(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=3,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=QINCo,
        num_quantizers=4,
        n_e=8192,
        e_dim=3,
        beta=0.25,
        top_a=64,
        hidden_dim=256,
        num_layers=3,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = QincoResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
                "top_a": top_a,
                "hidden_dim": hidden_dim,
                "num_layers": num_layers,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.qinco.f8_d4_e16384")
def QINCo_f8_d4_e16384(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=QINCo,
        num_quantizers=4,
        n_e=16384,
        e_dim=4,
        beta=0.25,
        top_a=64,
        hidden_dim=256,
        num_layers=3,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = QincoResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
                "top_a": top_a,
                "hidden_dim": hidden_dim,
                "num_layers": num_layers,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"discrete.qinco.f16_d8_e16384")
def QINCo_f16_d8_e16384(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=8,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=QINCo,
        num_quantizers=4,
        n_e=16384,
        e_dim=8,
        beta=0.25,
        top_a=64,
        hidden_dim=256,
        num_layers=3,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = QincoResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
                "top_a": top_a,
                "hidden_dim": hidden_dim,
                "num_layers": num_layers,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"continuous.soft_vq.f8_d16_e16384")
def SoftVQ_f8_d16_e16384(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=16,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=16,
        entropy_loss_weight=0.01,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        tau=0.07,
        use_norm=True,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    return VQModel(encoder, decoder, quantizer, **kwargs)

@register_model(f"continuous.soft_vq.f8_d32_e16384")
def SoftVQ_f8_d32_e16384(**kwargs):
    return SoftVQ_f8_d16_e16384(z_channels=32, e_dim=32, **kwargs)

@register_model(f"continuous.soft_vq.f8_d16_e16384_dinov2")
def SoftVQ_f8_d16_e16384_dinov2(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=16,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=16,
        entropy_loss_weight=0.01,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        tau=0.07,
        use_norm=True,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = VFFoundationAlignment(latent_channels=z_channels, foundation_type="dinov2")
    return VQModel(encoder, decoder, quantizer, alignment=alignment, **kwargs)

@register_model(f"continuous.soft_vq.f8_d32_e16384_dinov2")
def SoftVQ_f8_d32_e16384_dinov2(**kwargs):
    return SoftVQ_f8_d16_e16384_dinov2(z_channels=32, e_dim=32, **kwargs)

@register_model(f"continuous.soft_vq.f8_d16_e16384_biomedclip")
def SoftVQ_f8_d16_e16384_biomedclip(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=16,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[32],
        dropout=0.0,
        # --- quantizer config ---
        n_e=16384,
        e_dim=16,
        entropy_loss_weight=0.01,
        entropy_loss_temperature=0.01,
        entropy_gamma=1.0,
        tau=0.07,
        use_norm=True,
        **kwargs
    ):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,    
        use_norm=use_norm,
    )
    alignment = VFFoundationAlignment(latent_channels=z_channels, foundation_type="biomedclip")
    return VQModel(encoder, decoder, quantizer, alignment=alignment, **kwargs)


@register_model(f"continuous.soft_vq.f8_d32_e16384_biomedclip")
def SoftVQ_f8_d32_e16384_biomedclip(**kwargs):
    return SoftVQ_f8_d16_e16384_biomedclip(z_channels=32, e_dim=32, **kwargs)

@register_model(f"continuous.wqvae.f8_d4_e16384")
def WQVAE_f8_d4_e16384(
    # --- encoder/decoder config ---
        img_size=256,
        dims=2,
        double_z=False,
        z_channels=4,
        in_channels=3,
        out_ch=3,
        ch=128,
        ch_mult=[1, 2, 4],
        num_res_blocks=2,
        attn_resolutions=[],
        dropout=0.0,
        # --- quantizer config ---
        quantizer_class=WaveletResidualQuantizer,
        num_quantizers=4,
        n_e=16384,
        e_dim=4,
        beta=0.25,
        shared_codebook=True,
        quantize_dropout=False,
        dropout_start_level=0,
        wavelet="db1",
        wavelet_levels=1,
        subbands=["LL", "LH", "HL", "HH"],
        **kwargs
):
    encoder = Encoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=ch_mult,
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    decoder = Decoder(
        img_size=img_size,
        dims=dims,
        double_z=double_z,
        z_channels=z_channels,
        in_channels=in_channels,
        out_ch=out_ch,
        ch=ch,
        ch_mult=[ch_mult[0], ch_mult[1], ch_mult[2], ch_mult[2]],
        num_res_blocks=num_res_blocks,
        attn_resolutions=attn_resolutions,
        dropout=dropout
    )
    
    # --- Create identical quantizer kwargs for each level ---
    quantizer_kwargs_list = [
        {
            "n_e": n_e,
            "e_dim": e_dim,
            "beta": beta
        }
        for _ in range(num_quantizers)
    ]
    
    quantizer = quantizer_class(
        quantizer_class=VectorQuantizer,  # Base VQ class (your existing one)
        num_quantizers=num_quantizers,
        wavelet=wavelet,
        wavelet_levels=wavelet_levels,
        subbands=subbands,
        quantizer_kwargs_list=quantizer_kwargs_list,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    
    return VQModel(encoder, decoder, quantizer, **kwargs)


@register_model(f"discrete.hcvq.residual_vq.S_16")
def RQVAE_Transformer_S_16(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    mask_ratio_mu: float = 0.0,
    masking: str = "none",
    # --- quantizer config ---
    quantizer_class=VectorQuantizer2,
    num_quantizers=4,
    n_e=16384,
    e_dim=32,
    beta=0.25,
    shared_codebook=True,
    quantize_dropout=False,
    dropout_start_level=0,
    **kwargs
    ):

    encoder = GenericViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        mask_ratio_mu=mask_ratio_mu,
        masking=masking,
    )
    decoder = GenericViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        out_channels=in_channels,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        to_pixel="conv",
        token_dim=None,   # PostQuantLayer Is Done in the VQModel
    )
    quantizer = ResidualQuantizer(
        quantizer_class=quantizer_class,
        num_quantizers=num_quantizers,
        quantizer_kwargs_list=[
            {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
            }
        ] * num_quantizers,
        shared_codebook=shared_codebook,
        quantize_dropout=quantize_dropout,
        dropout_start_level=dropout_start_level,
    )
    return VQModelTransformer(encoder=encoder, decoder=decoder, quantizer=quantizer, pre_post_layer="linear", **kwargs)


@register_model(f"discrete.hcvq.soft_vq.S_16")
def SoftVQ_Transformer_S_16(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    mask_ratio_mu: float = 0.0,
    masking: str = "none",
    # --- quantizer config ---
    n_e=16384,
    e_dim=32,
    entropy_loss_weight=0.01,
    entropy_loss_temperature=0.01,
    entropy_gamma=1.0,
    tau=0.07,
    use_norm=True,
    **kwargs
    ):

    encoder = GenericViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        mask_ratio_mu=mask_ratio_mu,
        masking=masking,
    )
    decoder = GenericViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        out_channels=in_channels,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        to_pixel="conv",
        token_dim=None,   # PostQuantLayer Is Done in the VQModel
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    return VQModelTransformer(encoder=encoder, decoder=decoder, quantizer=quantizer, pre_post_layer="linear", **kwargs)


@register_model(f"discrete.hcvq.grouped_vq.S_16")
def GroupedVQ_Transformer_S_16(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    mask_ratio_mu: float = 0.0,
    masking: str = "none",
    # --- quantizer config ---
    quantizer_class=VectorQuantizer2,
    n_e=16384,
    e_dim=32,
    beta=0.25,
    groups=5,
    split_dim=1,
    **kwargs
    ):

    encoder = GenericViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        mask_ratio_mu=mask_ratio_mu,
        masking=masking,
    )
    decoder = GenericViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        out_channels=in_channels,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        to_pixel="conv",
        token_dim=None,   # PostQuantLayer Is Done in the VQModel
    )
    quantizer = GroupedVQ(
        quantizer_class=quantizer_class,
        quantizer_kwargs_list=[
                {
                "n_e": n_e,
                "e_dim": e_dim,
                "beta": beta,
                }
            ] * groups,
        groups=groups,
        split_dim=split_dim,
    )
    return VQModelTransformer(encoder=encoder, decoder=decoder, quantizer=quantizer, pre_post_layer="linear", **kwargs)


@register_model(f"discrete.hcvq.sd_vq.S_16")
def SDVQ_Transformer_S_16(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    mask_ratio_mu: float = 0.0,
    masking: str = "none",
    # --- quantizer config ---
    n_e=16384,
    e_dim=32,
    beta=0.25,
    use_ema=False,
    **kwargs
    ):

    encoder = GenericViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        mask_ratio_mu=mask_ratio_mu,
        masking=masking,
    )
    decoder = GenericViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        out_channels=in_channels,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        to_pixel="conv",
        token_dim=None,   # PostQuantLayer Is Done in the VQModel
    )
    quantizer = VectorQuantizer2(
        n_e=n_e,
        e_dim=e_dim,
        beta=beta,
        use_ema=use_ema,
    )
    return VQModelTransformer(encoder=encoder, decoder=decoder, quantizer=quantizer, pre_post_layer="linear", **kwargs)

@register_model(f"discrete.hcvq.msrq.S_16")
def MSRQ_Transformer_S_16(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    mask_ratio_mu: float = 0.0,
    masking: str = "none",
    # --- quantizer config ---
    dims=3,
    n_e=16384,
    e_dim=32,
    beta=0.25,
    use_ema=False,
    ema_decay=0.99,
    ema_eps=1e-5,
    v_patch_nums=((1,1,1), (2,2,2), (3,3,3), (4,4,4), (5,5,5), (6,6,6), (8,8,8), (10, 11, 9)),
    quant_resi=0.5,
    share_quant_resi=4,
    using_znorm=False,
    **kwargs
    ):

    encoder = GenericViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        mask_ratio_mu=mask_ratio_mu,
        masking=masking,
    )



    decoder = GenericViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        out_channels=in_channels,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        to_pixel="conv",
        token_dim=None,   # PostQuantLayer Is Done in the VQModel
    )
    

    quantizer = MultiScaleResidualQuantizer3D(
        n_e=n_e,
        e_dim=e_dim,
        dims=dims,
        using_znorm=using_znorm,
        beta=beta,
        use_ema=use_ema,
        ema_decay=ema_decay,
        ema_eps=ema_eps,
        v_patch_nums=v_patch_nums,
        quant_resi=quant_resi,
        share_quant_resi=share_quant_resi,
    )
    return VQModelTransformer(encoder=encoder, decoder=decoder, quantizer=quantizer, pre_post_layer="linear", **kwargs)


import logging
import random
from functools import partial

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from einops.layers.torch import Rearrange
from torch import Tensor
from medlat.modules.in_and_out import PatchEmbed, ToPixel
from medlat.modules.pos_embed import get_rope_tensor_2d, get_rope_tensor_3d, apply_rotary_emb

from medlat.first_stage.modules.gaussian_dist import DiagonalGaussianDistribution
from medlat.utils import init_from_ckpt
from medlat.registry import register_model

logger = logging.getLogger("DeTok")



# ================================
# Utility Functions
# ================================

def _to_tensor(x):
    return x.clone().detach() if isinstance(x, torch.Tensor) else torch.tensor(x)


# ================================
# Neural Network Components
# ================================


class SwiGLUFFN(nn.Module):
    """Swish-Gated Linear Unit Feed-Forward Network."""

    def __init__(self, in_features: int, hidden_features: int = None, out_features: int = None):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.w12 = nn.Linear(in_features, 2 * hidden_features)
        self.w3 = nn.Linear(hidden_features, out_features)

    def forward(self, x: Tensor) -> Tensor:
        x1, x2 = self.w12(x).chunk(2, dim=-1)
        return self.w3(F.silu(x1) * x2)


class Attention(nn.Module):
    """multi-head attention with optional rotary position embedding."""

    def __init__(self, dim: int, num_heads: int = 8, use_rope: bool = True) -> None:
        super().__init__()
        assert dim % num_heads == 0, f"dim % num_heads !=0, got {dim} and {num_heads}"
        self.head_dim = dim // num_heads
        self.num_heads = num_heads
        self.use_rope = use_rope
        self.qkv = nn.Linear(dim, dim * 3)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x: Tensor, rope: Tensor = None) -> Tensor:
        bsz, n_ctx, ch = x.shape
        qkv = self.qkv(x)
        q, k, v = rearrange(qkv, "b n (qkv h d) -> qkv b h n d", qkv=3, h=self.num_heads).unbind(0)
        
        if self.use_rope and rope is not None:
            q, k = apply_rotary_emb(q, rope), apply_rotary_emb(k, rope)
        
        x = F.scaled_dot_product_attention(q, k, v)
        return self.proj(x.transpose(1, 2).reshape(bsz, n_ctx, ch))


class Block(nn.Module):
    """transformer block with attention and feed-forward layers."""

    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        norm_layer=partial(nn.RMSNorm, eps=1e-6),
        use_rope: bool = True,
    ) -> None:
        super().__init__()
        self.norm1, self.norm2 = norm_layer(dim), norm_layer(dim)
        self.attn = Attention(dim, num_heads, use_rope=use_rope)
        self.mlp = SwiGLUFFN(dim, int(2 / 3 * dim * mlp_ratio))

    def forward(self, x: Tensor, rope: Tensor = None) -> Tensor:
        x = x + self.attn(self.norm1(x), rope=rope)
        x = x + self.mlp(self.norm2(x))
        return x


# ================================
# Encoder and Decoder
# ================================


class Encoder(nn.Module):
    """vision Transformer encoder with masked autoencoding capability."""

    def __init__(
        self,
        img_size: int | tuple[int, ...] = 256,
        patch_size: int | tuple[int, ...] = 16,
        in_channels: int = 3,
        width: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        token_channels: int = 16,
        mask_ratio: float = 0.75,
        dims: int = 2,
        use_rope: bool = True,
    ) -> None:
        super().__init__()
        self.img_size = img_size
        self.patch_size = patch_size
        self.dims = dims
        # needs to split into mean and std
        self.token_channels = token_channels * 2
        self.mask_ratio = mask_ratio
        self.in_channels = in_channels
        
        # Handle different input formats
        if isinstance(img_size, int):
            if dims == 2:
                self.img_size = (img_size, img_size)
            elif dims == 3:
                self.img_size = (img_size, img_size, img_size)
        else:
            self.img_size = img_size
            
        if isinstance(patch_size, int):
            if dims == 2:
                self.patch_size = (patch_size, patch_size)
            elif dims == 3:
                self.patch_size = (patch_size, patch_size, patch_size)
        else:
            self.patch_size = patch_size
            
        # Calculate grid sizes
        if dims == 2:
            self.grid_size = (self.img_size[0] // self.patch_size[0], self.img_size[1] // self.patch_size[1])
            self.seq_len = self.grid_size[0] * self.grid_size[1]
        elif dims == 3:
            self.grid_size = (self.img_size[0] // self.patch_size[0], 
                            self.img_size[1] // self.patch_size[1], 
                            self.img_size[2] // self.patch_size[2])
            self.seq_len = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
        else:
            raise ValueError(f"Unsupported dims: {dims}")

        num_layers, num_heads, width = depth, num_heads, width
        
        self.width = width

        # Patch embedding
        self.patch_embed = PatchEmbed(to_embed='conv', img_size=img_size, patch_size=patch_size, 
                                     in_chans=in_channels, embed_dim=width)

        # learnable embeddings
        scale = width**-0.5
        self.positional_embedding = nn.Parameter(scale * torch.randn(1, self.seq_len, width))

        # transformer layers
        norm_layer = partial(nn.RMSNorm, eps=1e-6)
        self.ln_pre = norm_layer(width)
        self.transformer = nn.ModuleList(
            [Block(dim=width, num_heads=num_heads, norm_layer=norm_layer, use_rope=use_rope) for _ in range(num_layers)]
        )
        self.ln_post = norm_layer(width)
        self.latent_head = nn.Linear(width, self.token_channels)

        # rotary position embedding (only for 2D)
        if dims == 2 and use_rope:
            head_dim = self.transformer[0].attn.head_dim
            rope_tensor = get_rope_tensor_2d(head_dim, self.grid_size[0], self.grid_size[1]).unsqueeze(0)
            self.register_buffer("rope_tensor", rope_tensor, persistent=False)
        elif dims == 3 and use_rope:
            head_dim = self.transformer[0].attn.head_dim
            rope_tensor = get_rope_tensor_3d(head_dim, self.grid_size[0], self.grid_size[1], self.grid_size[2]).unsqueeze(0)
            self.register_buffer("rope_tensor", rope_tensor, persistent=False)
        else:
            self.register_buffer("rope_tensor", None, persistent=False)



    def mae_random_masking(self, x: Tensor, mask_ratio: float = -1):
        """apply masked autoencoding random masking."""
        bsz, seq_len, chans = x.shape
        # mask: 0 for visible, 1 for masked
        if mask_ratio == 0:
            # no masking
            rope = self.rope_tensor.expand(bsz, -1, -1)
            return x, torch.zeros(bsz, seq_len, device=x.device), None, rope

        if mask_ratio < 0:
            mask_ratio = max(0.0, random.uniform(-0.1, self.mask_ratio))
        len_keep = int(np.ceil(seq_len * (1 - mask_ratio)))
        noise = torch.rand(bsz, seq_len, device=x.device)
        ids_shuffle = torch.argsort(noise, dim=1)
        ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :len_keep]
        x_visible = torch.gather(x, 1, ids_keep[..., None].repeat(1, 1, chans))
        rope = self.rope_tensor.expand(bsz, -1, -1)
        rope_visible = torch.gather(rope, 1, ids_keep[..., None].repeat(1, 1, rope.shape[-1]))

        mask = torch.ones(bsz, seq_len, device=x.device)
        mask[:, :len_keep] = 0
        mask = torch.gather(mask, dim=1, index=ids_restore)
        return x_visible, mask, ids_restore, rope_visible

    def forward(self, x: Tensor, mask_ratio: float = -1):
        """forward pass through encoder."""
        x = self.patch_embed(x)

        x = x + self.positional_embedding
        x, _, ids_restore, rope = self.mae_random_masking(x, mask_ratio=mask_ratio)

        x = self.ln_pre(x)
        for block in self.transformer:
            x = block(x, rope)
        x = self.ln_post(x)

        tokens = self.latent_head(x)

        return tokens, ids_restore


class Decoder(nn.Module):
    """vision Transformer decoder with mask tokens for image reconstruction."""

    def __init__(
        self,
        img_size: int | tuple[int, ...] = 256,
        patch_size: int | tuple[int, ...] = 16,
        out_channels: int = 3,
        width: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        token_channels: int = 16,
        dims: int = 2,
        use_rope: bool = True,
    ) -> None:
        super().__init__()
        self.img_size = img_size
        self.patch_size = patch_size
        self.dims = dims
        self.width = width
        self.depth = depth
        self.num_heads = num_heads
        self.token_channels = token_channels
        self.out_channels = out_channels
        # Handle different input formats
        if isinstance(img_size, int):
            if dims == 2:
                self.img_size = (img_size, img_size)
            elif dims == 3:
                self.img_size = (img_size, img_size, img_size)
        else:
            self.img_size = img_size
            
        if isinstance(patch_size, int):
            if dims == 2:
                self.patch_size = (patch_size, patch_size)
            elif dims == 3:
                self.patch_size = (patch_size, patch_size, patch_size)
        else:
            self.patch_size = patch_size
        
        # Calculate grid sizes
        if dims == 2:
            self.grid_size = (self.img_size[0] // self.patch_size[0], self.img_size[1] // self.patch_size[1])
            self.seq_len = self.grid_size[0] * self.grid_size[1]
            self.output_channels = 3
        elif dims == 3:
            self.grid_size = (self.img_size[0] // self.patch_size[0], 
                            self.img_size[1] // self.patch_size[1], 
                            self.img_size[2] // self.patch_size[2])
            self.seq_len = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
            self.output_channels = 1  # For 3D, typically single channel
        else:
            raise ValueError(f"Unsupported dims: {dims}")

        num_layers, num_heads, width = depth, num_heads, width

        # learnable embeddings
        scale = width**-0.5
        self.positional_embedding = nn.Parameter(scale * torch.randn(1, self.seq_len, width))
        self.mask_token = nn.Parameter(scale * torch.randn(1, 1, width))

        # decoder layers
        self.decoder_embed = nn.Linear(self.token_channels, width)
        norm_layer = partial(nn.RMSNorm, eps=1e-6)
        self.ln_pre = norm_layer(width)
        self.transformer = nn.ModuleList(
            [Block(dim=width, num_heads=num_heads, norm_layer=norm_layer, use_rope=use_rope) for _ in range(num_layers)]
        )
        self.ln_post = norm_layer(width)

        # output layers
        # To pixel head
        self.to_pixel = ToPixel(to_pixel="conv", img_size=self.img_size, out_channels=out_channels,
                               in_dim=width, patch_size=self.patch_size)
        # rotary position embedding
        if dims == 2 and use_rope:
            head_dim = self.transformer[0].attn.head_dim
            rope_tensor = get_rope_tensor_2d(head_dim, self.grid_size[0], self.grid_size[1]).unsqueeze(0)
            self.register_buffer("rope_tensor", rope_tensor, persistent=False)
        elif dims == 3 and use_rope:
            head_dim = self.transformer[0].attn.head_dim
            rope_tensor = get_rope_tensor_3d(head_dim, self.grid_size[0], self.grid_size[1], self.grid_size[2]).unsqueeze(0)
            print(rope_tensor.shape)
            self.register_buffer("rope_tensor", rope_tensor, persistent=False)
        else:
            self.register_buffer("rope_tensor", None, persistent=False)

    def forward(self, z_latents: Tensor, ids_restore: Tensor | None = None) -> Tensor:
        """forward pass through decoder."""
        z = self.decoder_embed(z_latents)
        bsz, seq_len, _ = z.shape

        if ids_restore is not None:
            num_mask_tokens = ids_restore.shape[1] + 1 - seq_len
            mask_tokens = self.mask_token.repeat(bsz, num_mask_tokens, 1)
            z_ = torch.cat([z, mask_tokens], dim=1)
            expanded_ids_restore = ids_restore.unsqueeze(-1).expand(-1, -1, z_.shape[-1])
            z = torch.gather(z_, dim=1, index=expanded_ids_restore)

        z = z + self.positional_embedding

        z = self.ln_pre(z)
        if self.rope_tensor is not None:
            rope = self.rope_tensor.expand(bsz, -1, -1)
        else:
            rope = None
        for block in self.transformer:
            z = block(z, rope)
        z = self.ln_post(z)

        z = self.to_pixel(z)

        return z


# ================================
# Main DeTok Model
# ================================


class DeTok(nn.Module): 
    """
    l-DeTok: latent denoising makes good visual tokenizers.
    Supports both 2D and 3D inputs with arbitrary dimss.
    """
    def __init__(
        self,
        img_size: int | tuple[int, ...] = 256,
        patch_size: int | tuple[int, ...] = 16,
        in_channels: int = 3,
        out_channels: int = 3,
        enc_width: int = 768,
        dec_width: int = 768,
        enc_depth: int = 12,
        dec_depth: int = 12,
        enc_num_heads: int = 12,
        dec_num_heads: int = 12,
        token_channels: int = 16,
        mask_ratio: float = 0.75,
        gamma: float = 3.0,
        use_additive_noise: bool = False,
        dims: int = 2,
        # normalization parameters used for generative model training
        mean=0.0,
        std=1.0,
        scale_factor: float = 1.0,
        ckpt_path: str = None,
        use_rope: bool = True,
        kl_weight: float = 1e-6,
    ) -> None:
        super().__init__()

        # initialize encoder and decoder
        self.encoder = Encoder(
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            width=enc_width,
            depth=enc_depth,
            num_heads=enc_num_heads,
            token_channels=token_channels,
            mask_ratio=mask_ratio,
            dims=dims,
            use_rope=use_rope,
        )
        self.decoder = Decoder(
            img_size=img_size,
            patch_size=patch_size,
            out_channels=out_channels,
            width=dec_width,
            depth=dec_depth,
            num_heads=dec_num_heads,
            token_channels=token_channels,
            dims=dims,
            use_rope=use_rope,
        )

        # model configuration
        self.dims = dims
        self.img_size = img_size
        self.patch_size = patch_size
        self.kl_weight = kl_weight
        
        # Calculate grid sizes for tokenization
        if isinstance(img_size, int):
            if dims == 2:
                self.img_size = (img_size, img_size)
            elif dims == 3:
                self.img_size = (img_size, img_size, img_size)
        else:
            self.img_size = img_size
            
        if isinstance(patch_size, int):
            if dims == 2:
                self.patch_size_tuple = (patch_size, patch_size)
            elif dims == 3:
                self.patch_size_tuple = (patch_size, patch_size, patch_size)
        else:
            self.patch_size_tuple = patch_size
            
        if dims == 2:
            self.grid_size = (self.img_size[0] // self.patch_size_tuple[0], self.img_size[1] // self.patch_size_tuple[1])
        elif dims == 3:
            self.grid_size = (self.img_size[0] // self.patch_size_tuple[0], 
                            self.img_size[1] // self.patch_size_tuple[1], 
                            self.img_size[2] // self.patch_size_tuple[2])
            
        self.width = enc_width
        self.use_additive_noise = use_additive_noise
        self.gamma = gamma

        self.scale_factor = scale_factor

        # initialize weights
        self.apply(self._init_weights)

        # setup to-posteriors function
        self.to_posteriors = partial(DiagonalGaussianDistribution, channel_dim=-1)

        # setup normalization parameters
        if isinstance(mean, np.ndarray) or isinstance(mean, list):
            mean = np.array(mean).reshape(1, -1, 1, 1)
            std = np.array(std).reshape(1, -1, 1, 1)
        self.register_buffer("mean", torch.tensor(mean), persistent=False)
        self.register_buffer("std", torch.tensor(std), persistent=False)

        params_M = sum(p.numel() for p in self.parameters() if p.requires_grad) / 1e6
        logger.info(f"[DeTok] params: {params_M:.2f}M, {dims}D, size: {self.img_size}, patch: {self.patch_size_tuple}")
        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)

    def _init_weights(self, module: nn.Module) -> None:
        """initialize the weights."""
        if isinstance(module, (nn.Linear, nn.Conv1d, nn.Conv2d, nn.Conv3d)):
            module.weight.data = nn.init.trunc_normal_(module.weight.data, mean=0.0, std=0.02)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data = nn.init.trunc_normal_(module.weight.data, mean=0.0, std=0.02)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def freeze_everything_but_decoder(self) -> None:
        """freeze all parameters except the decoder, used for decoder fine-tuning"""
        for param in self.parameters():
            param.requires_grad = False

        for param in self.decoder.parameters():
            param.requires_grad = True

        params_M = sum(p.numel() for p in self.parameters() if p.requires_grad) / 1e6
        logger.info(f"[DeTok] trainable params: {params_M:.2f}M (after freezing all but decoder)")

    def reset_stats(self, mean: Tensor | np.ndarray | float, std: Tensor | np.ndarray | float) -> None:
        if isinstance(mean, float) and isinstance(std, float) or (mean.ndim == 0 and std.ndim == 0):
            # a single digit global mean and global std
            self.register_buffer("mean", _to_tensor(mean), persistent=False)
            self.register_buffer("std", _to_tensor(std), persistent=False)
        else:
            n_chans = mean.shape[-1]
            self.register_buffer("mean", _to_tensor(mean).reshape(1, 1, n_chans), persistent=False)
            self.register_buffer("std", _to_tensor(std).reshape(1, 1, n_chans), persistent=False)
        logger.info(f"Resetting mean and std ({mean.shape=}, {std.shape=})")
        logger.info(f"Mean: {self.mean}")
        logger.info(f"Std: {self.std}")

    def denormalize_z(self, z: Tensor) -> Tensor:
        """denormalize latent tokens."""
        return z * self.std.to(z) / self.scale_factor + self.mean.to(z)

    def normalize_z(self, z: Tensor) -> Tensor:
        """normalize latent tokens."""
        return (z - self.mean.to(z)) * self.scale_factor / self.std.to(z)

    def encode_into_posteriors(self, x: Tensor):
        """encode image into posterior distributions."""
        z = self.encoder(x, mask_ratio=0.0)[0]
        return self.to_posteriors(z)

    def encode(self, x: Tensor, sampling: bool = False, mask_ratio: float = -1, noise_level: float = -1.0):
        """encode image into latent tokens."""
        z, ids_restore = self.encoder(x, mask_ratio=mask_ratio)

        posteriors = self.to_posteriors(z)
        z_latents = posteriors.sample() if sampling else posteriors.mean

        if self.training and self.gamma > 0.0:
            device = z_latents.device
            bsz, n_tokens, chans = z_latents.shape
            if noise_level > 0.0:
                noise_level_tensor = torch.full((bsz, 1, 1), noise_level, device=device)
            else:
                noise_level_tensor = torch.rand(bsz, 1, 1, device=device)
            noise_level_tensor = noise_level_tensor.expand(-1, n_tokens, chans)
            noise = torch.randn(bsz, n_tokens, chans, device=device) * self.gamma
            if self.use_additive_noise:
                z_latents = z_latents + noise_level_tensor * noise
            else:
                z_latents = (1 - noise_level_tensor) * z_latents + noise_level_tensor * noise

        return z_latents, posteriors, ids_restore

    def p_loss(self, posterior, device): 
        kl_loss = torch.zeros((), device=device)
        if posterior is not None:
            # assume extra_result_dict contains posteriors with kl method
            if hasattr(posterior, "kl"):
                kl_loss = posterior.kl()
                kl_loss = torch.sum(kl_loss) / kl_loss.shape[0]
        kl_loss = self.kl_weight * kl_loss
        return kl_loss

    def forward(self, x: Tensor):
        """forward pass through the entire model."""
        z_latents, posteriors, ids_restore = self.encode(x, sampling=self.training)
        decoded = self.decoder(z_latents, ids_restore=ids_restore)
        return decoded, self.p_loss(posteriors, x.device)

    def tokenize(self, x: Tensor, sampling: bool = False) -> Tensor:
        """tokenize input image and normalize the latent tokens."""
        z = self.encode(x, sampling=sampling, mask_ratio=0.0)[0]
        z = self.normalize_z(z)
        if self.dims == 2:
            z = rearrange(z, "b (h w) c -> b c h w", h=self.grid_size[0], w=self.grid_size[1])
        elif self.dims == 3:
            z = rearrange(z, "b (d h w) c -> b c d h w", d=self.grid_size[0], h=self.grid_size[1], w=self.grid_size[2])
        return z

    def detokenize(self, z: Tensor) -> Tensor:
        """detokenize latent representation back to image."""
        if self.dims == 2:
            z = rearrange(z, "b c h w -> b (h w) c")
        elif self.dims == 3:
            z = rearrange(z, "b c d h w -> b (d h w) c")
        z = self.denormalize_z(z)
        decoded_images = self.decoder(z)
        return decoded_images

    def sample_from_moments(self, moments: Tensor) -> Tensor:
        """sample from latent moments."""
        z = DiagonalGaussianDistribution(moments, channel_dim=-1).sample()
        z = self.normalize_z(z)
        if self.dims == 2:
            z = rearrange(z, "b (h w) c -> b c h w", h=self.grid_size[0], w=self.grid_size[1])
        elif self.dims == 3:
            z = rearrange(z, "b (d h w) c -> b c d h w", d=self.grid_size[0], h=self.grid_size[1], w=self.grid_size[2])
        return z

    @torch.inference_mode()
    def reconstruct(self, x: Tensor) -> Tensor:
        """reconstruct input image."""
        return self.detokenize(self.tokenize(x))


# ================================
# Model Factory Functions
# ================================

@register_model("token.detok.ss")
def detok_SS(**kwargs) -> DeTok:
    return DeTok(enc_width=512, dec_width=512, enc_depth=8, dec_depth=8, enc_num_heads=8, dec_num_heads=8, **kwargs)


@register_model("token.detok.sb")
def detok_SB(**kwargs) -> DeTok:
    return DeTok(enc_width=512, dec_width=768, enc_depth=8, dec_depth=12, enc_num_heads=8, dec_num_heads=12, **kwargs)


@register_model("token.detok.sl")
def detok_SL(**kwargs) -> DeTok:
    return DeTok(enc_width=512, dec_width=1024, enc_depth=8, dec_depth=24, enc_num_heads=8, dec_num_heads=16, **kwargs)


@register_model("token.detok.bs")
def detok_BS(**kwargs) -> DeTok:
    return DeTok(enc_width=768, dec_width=512, enc_depth=12, dec_depth=8, enc_num_heads=12, dec_num_heads=8, **kwargs)


@register_model("token.detok.bb")
def detok_BB(**kwargs) -> DeTok:
    return DeTok(enc_width=768, dec_width=768, enc_depth=12, dec_depth=12, enc_num_heads=12, dec_num_heads=12, **kwargs)


@register_model("token.detok.bl")
def detok_BL(**kwargs) -> DeTok:
    return DeTok(enc_width=768, dec_width=1024, enc_depth=12, dec_depth=24, enc_num_heads=12, dec_num_heads=16, **kwargs)


@register_model("token.detok.ls")
def detok_LS(**kwargs) -> DeTok:
    return DeTok(enc_width=1024, dec_width=512, enc_depth=24, dec_depth=8, enc_num_heads=16, dec_num_heads=8, **kwargs)


@register_model("token.detok.lb")
def detok_LB(**kwargs) -> DeTok:
    return DeTok(enc_width=1024, dec_width=768, enc_depth=24, dec_depth=12, enc_num_heads=16, dec_num_heads=12, **kwargs)


@register_model("token.detok.ll")
def detok_LL(**kwargs) -> DeTok:
    return DeTok(enc_width=1024, dec_width=1024, enc_depth=24, dec_depth=24, enc_num_heads=16, dec_num_heads=16, **kwargs)


@register_model("token.detok.xlxl")
def detok_XLXL(**kwargs) -> DeTok:
    return DeTok(enc_width=1152, dec_width=1152, enc_depth=28, dec_depth=28, enc_num_heads=16, dec_num_heads=16, **kwargs)





from typing import Optional, Tuple
import torch
import torch.nn.functional as F
import numpy as np
import torch.nn as nn


def sincos_pos_embed(embed_dim, grid_size, cls_token=False, use_both_axes=True, circular_pe=False, shift_patch: Optional[tuple] = (0, 0, 0)):
    """
    Generate multi-scale sine-cosine position embedding for 3D+t images with both long axis and short axis slices.
    
    grid_size: int of the grid (S, T, H/patch_size, W/patch_size) where the first 3 slices should be long axis slices
    embed_dim: output dimension for each position where the first dimension is to distinguish long axis from short axis
    pos_embed: [np.prod(grid_size), embed_dim] or [1+np.prod(grid_size), embed_dim] (w/ or w/o cls_token)
    circular_pe: boolean, indicating whether the positional embedding of temporal dimension setup as circular or not.
    shift_patch: the number of the patches to be slided of shape (t_slide, h_slide, w_slide)
    """
    assert len(shift_patch) == 3
    _shift_patch = (0,) + shift_patch
    if use_both_axes:
        assert len(grid_size) >= 3, "Grid_size should be at least 3D for positional embeding with long axis"
        assert (embed_dim - 1) % (len(grid_size) * 2) == 0, "Each dimension has 2 channels (sin, cos)"
        grid_size_sax = (grid_size[0] - 3, *grid_size[1:])
        grid_size_lax = (3, *grid_size[1:])
        
        if not circular_pe:
            grid_lax = torch.meshgrid(*[torch.arange(s, e+s, dtype=torch.float32) for (s, e) in zip(_shift_patch, grid_size_lax)], indexing='ij')
            grid_sax = torch.meshgrid(*[torch.arange(s, e+s, dtype=torch.float32) for (s, e) in zip(_shift_patch, grid_size_sax)], indexing='ij')
        else:
            grid_lax, grid_sax = [], []
            for i, (s, e_l, e_s) in enumerate(zip(_shift_patch, grid_size_lax, grid_size_sax)):
                if i == 1: # temporal
                    _g_lax = torch.arange(s, e_l+s, dtype=torch.float32)
                    _g_sax = torch.arange(s, e_s+s, dtype=torch.float32)
                    g_lax = torch.sin(_g_lax * 2 * torch.pi / len(_g_lax))
                    g_sax = torch.sin(_g_sax * 2 * torch.pi / len(_g_sax))
                else:
                    g_lax = torch.arange(s, e_l+s, dtype=torch.float32)
                    g_sax = torch.arange(s, e_s+s, dtype=torch.float32)
                grid_lax.append(g_lax)
                grid_sax.append(g_sax)
            grid_lax = torch.meshgrid(grid_lax, indexing='ij')
            grid_sax = torch.meshgrid(grid_sax, indexing='ij')
        _grid_lax = torch.stack(grid_lax, dim=0)
        _grid_sax = torch.stack(grid_sax, dim=0)
        _pe_lax = get_multi_sincos_pos_embed_from_grid(embed_dim - 1, _grid_lax)
        _pe_sax = get_multi_sincos_pos_embed_from_grid(embed_dim - 1, _grid_sax)
        _pe_axes = torch.cat([torch.zeros([_pe_lax.shape[0], 1]), torch.ones([_pe_sax.shape[0], 1])], dim=0)
        _pe = torch.cat([_pe_lax, _pe_sax], dim=0)
        pos_embed = torch.cat([_pe_axes, _pe], dim=1)
    else:
        assert len(grid_size) >= 2, "Grid_size should be at least 2D"
        assert embed_dim % (len(grid_size) * 2) == 0, "Each dimension has 2 channels (sin, cos)"
        if not circular_pe:
            grid = torch.meshgrid(*[torch.arange(s, e-s, dtype=torch.float32) for (s, e) in zip(_shift_patch, grid_size)], indexing='ij')
        else:
            grid = []
            for i, (s, e) in enumerate(zip(_shift_patch, grid_size)):
                if i == 1: # temporal
                    _g = torch.arange(s, e_l+s, dtype=torch.float32)
                    g = torch.sin(_g * 2 * torch.pi / len(_g))
                else:
                    g = torch.arange(s, e+s, dtype=torch.float32)
                grid.append(g)
            grid = torch.meshgrid(grid, indexing='ij')
        grid = torch.stack(grid, dim=0)
        pos_embed = get_multi_sincos_pos_embed_from_grid(embed_dim, grid)
        
    if cls_token:
        pos_embed = torch.concatenate([torch.zeros([1, embed_dim]), pos_embed], dim=0)
    return pos_embed


def get_multi_sincos_pos_embed_from_grid(embed_dim, grid):
    # use half of dimensions to encode grid
    grid_dim = len(grid.shape) - 1
    emb = [get_1d_sincos_pos_embed_from_grid(embed_dim // grid_dim, grid[i]) for i in range(grid.shape[0])]
    emb = torch.concatenate(emb, dim=1) # [(S*T*H*W, D/4)] -> (S*T*H*W, D)
    return emb


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = torch.arange(embed_dim // 2, dtype=torch.float)
    omega /= embed_dim / 2.
    omega = 1. / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = torch.einsum("m,d->md", pos, omega)  # (M, D/2), outer product

    emb_sin = torch.sin(out) # (M, D/2)
    emb_cos = torch.cos(out) # (M, D/2)

    emb = torch.concatenate([emb_sin, emb_cos], dim=1)  # (M, D)
    return emb

 
def patchify(im: torch.Tensor, patch_size: list[int], **kwargs):
    """Split image into patches of size patch_size.
    
    im: [B, S, T, H, W]
    patch_size: a list of 3
    x: [B, L, np.prod(patch_size)] where L = S * T * H * W / np.prod(patch_size)
    """
    assert len(im.shape) == 5
    B, S, T, H, W = im.shape
    
    if len(patch_size) == 3:
        t, h, w = T // patch_size[0], H // patch_size[1], W // patch_size[2]
        x = im.reshape(B, S, t, patch_size[0], h, patch_size[1], w, patch_size[2])
        x = torch.einsum("bstphqwr->bsthwpqr", x)
        x = x.reshape(B, S * t * h * w, np.prod(patch_size))
    elif len(patch_size) == 2:
        h, w = H // patch_size[0], W // patch_size[1]
        x = im.reshape(B, S, T, h, patch_size[0], w, patch_size[1])
        x = torch.einsum("bsthqwr->bshwtqr", x)
        x = x.reshape(B, S * h * w, T * np.prod(patch_size))
    return x


def unpatchify(x: torch.Tensor, im_shape: list[int], patch_size: list[int, int, int]=[5, 16, 16], **kwargs):
    """Combine patches into image.
    
    x: [B, L, np.prod(patch_size) or T * np.prod(patch_size)]
    im_shape: [B, S, T, X, Y]
    im: [B, S, T, X, Y] where X = Y
    """
    assert len(x.shape) == 3
    assert len(im_shape) == 5
    
    B, S, T, H, W = im_shape
    if len(patch_size) == 3:
        t, h, w = T // patch_size[0], H // patch_size[1], W // patch_size[2]
        x = x.reshape(B, S, t, h, w, patch_size[0], patch_size[1], patch_size[2])
        x = torch.einsum("bsthwpqr->bstphqwr", x)
        im = x.reshape(im_shape)
    elif len(patch_size) == 2:
        h, w = H // patch_size[0], W // patch_size[1]
        x = x.reshape(B, S, h, w, T, patch_size[0], patch_size[1])
        x = torch.einsum("bshwtqr->bsthqwr", x)
        im = x.reshape(im_shape)
    return im


def patchify_SAX(im: torch.Tensor, patch_size: list[int], S_sax: int, in_channels: int = 6, pixel_unshuffle_scale: int = 1, **kwargs):
    """Split image into patches of size patch_size.
    
    im: [B, S, T, H, W]
    patch_size: a list of 3
    x: [B, L, np.prod(patch_size)] where L = S * T * H * W / np.prod(patch_size)
    """
    assert len(im.shape) == 5
    assert len(patch_size) == 3
    B, S, T, H, W = im.shape
    if S == S_sax:
        x = im[:, None] # (B, 1, S, T, H, W)
        S_new = 1
    elif S > S_sax:
        S_lax = S - S_sax
        S_new = 1 + S_lax
        
        sax = im[:, 3:, ...] # (B, S_lax, T, H, W)
        lax = im[:, :3, ...] # (B, S_lax, T, H, W)
        sax_ = sax[:, None] # (B, 1, S_sax, T, H, W)
        lax_ = lax[:, :, None] # (B, S_lax, 1, T, H, W)
        lax_rep = torch.tile(lax_, dims=(1, 1, S_sax, 1, 1, 1)) # (B, S_lax, S_sax, T, H, W)
        x = torch.cat([lax_rep, sax_], dim=1) # (B, S_lax + 1, S_sax, T, H, W)
    else:
        S_new = im.shape[1]
        x = torch.tile(im[:, :, None], dims=(1, 1, in_channels, 1, 1, 1)) # (B, S_lax, in_channels, T, H, W)
    if pixel_unshuffle_scale != 1:
        x = pixel_unshuffle3d(x.reshape(-1, *x.shape[2:]), pixel_unshuffle_scale) # (B*S_new, S_sax*r**3, T//r, H//r, W//r)
        x = x.reshape(B, -1, *x.shape[1:])
    t = T // patch_size[0] // pixel_unshuffle_scale
    h = H // patch_size[1] // pixel_unshuffle_scale
    w = W // patch_size[2] // pixel_unshuffle_scale
    x = x.reshape(B, S_new, in_channels, t, patch_size[0], h, patch_size[1], w, patch_size[2])
    x = torch.einsum("bsitphqwr->bsthwipqr", x)
    x = x.reshape(B, S_new * t * h * w, in_channels * np.prod(patch_size))

    return x


def unpatchify_SAX(x: torch.Tensor, im_shape: list[int], patch_size: list[int, int, int], 
                   S_sax: int, in_channels: int = 6, pixel_unshuffle_scale: int = 1, **kwargs):
    """Combine patches into image.
    
    x: [B, L, np.prod(patch_size) or T * np.prod(patch_size)]
    im_shape: [B, S, T, X, Y]
    im: [B, S, T, X, Y] where X = Y
    """
    assert len(x.shape) == 3
    assert len(im_shape) == 5

    B, S, T, H, W = im_shape
    t = T // patch_size[0] // pixel_unshuffle_scale
    h = H // patch_size[1] // pixel_unshuffle_scale
    w = W // patch_size[2] // pixel_unshuffle_scale
    if S == S_sax:
        x = x.reshape(B, 1, t, h, w, in_channels, patch_size[0], patch_size[1], patch_size[2])
        x = torch.einsum("bsthwipqr->bsitphqwr", x) # (B, S_new, channel)
        x = x.reshape(B, 1, in_channels, T, H, W)
        im = x[:, 0, ...]
    elif S > S_sax:
        S_lax = S - S_sax
        S_new = 1 + S_lax
        
        x = x.reshape(B, S_new, t, h, w, in_channels, patch_size[0], patch_size[1], patch_size[2])
        x = torch.einsum("bsthwipqr->bsitphqwr", x) # (B, S_new, channel)
        x = x.reshape(B, S_new, in_channels, t*patch_size[0], h*patch_size[1], w*patch_size[2])
        if pixel_unshuffle_scale != 1:
            x = pixel_shuffle3d(x.reshape(-1, *x.shape[2:]), pixel_unshuffle_scale)
            x = x.reshape(B, -1, *x.shape[1:])
        lax = x[:, :3, 0] # (B, S_lax, T, H, W)
        sax = x[:, -1] # (B, S_lax, T, H, W)
        im = torch.cat([lax, sax], dim=1) # (B, S, T, H, W)
    else:
        x = x.reshape(B, im_shape[1], t, h, w, in_channels, patch_size[0], patch_size[1], patch_size[2])
        x = torch.einsum("bsthwipqr->bsitphqwr", x) # (B, S_new, channel)
        x = x.reshape(B, im_shape[1], in_channels, T, H, W)
        im = x[:, :, 0, ...]
    return im

    
class Masker:
    def __init__(self, mask_type, mask_ratio, grid_size, **kwargs):
        self.mask_ratio = mask_ratio
        if mask_type == "random":
            self.masking_strategy = self.random_masking
        else:
            raise NotImplementedError
    
    def __call__(self, x):
        """
        x: [N, L, D], sequence
        x_masked: [N, L * mask_ratio, D], masked sequence
        mask: [N, L], binary mask
        ids_restore: [N, L], indices to restore the original order
        """
        mask, ids_restore, ids_keep = self.masking_strategy(input_size=x.shape, device=x.device)
        x_masked = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.shape[-1]))
        return x_masked, mask, ids_restore
    
    def call_masking_fctn(self, x, fctn_name, **kwargs):
        fctn = eval(f"self.{fctn_name}")
        mask, ids_restore, ids_keep = fctn(input_size=x.shape, device=x.device, **kwargs)
        x_masked = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.shape[-1]))
        return x_masked, mask, ids_restore
        
    def random_masking(self, input_size, device, **kwargs):
        """
        # Reference: https://github.com/facebookresearch/mae.git
        
        Perform per-sample random masking by per-sample shuffling.
        Per-sample shuffling is done by argsort random noise.
        
        input_size: [N, L, D], sequence
        device: torch.device
        mask: [N, L], binary mask
        ids_restore: [N, L], indices to restore the original order
        """
        N, L, D = input_size  # batch, length, dim
        len_keep = int(L * (1 - self.mask_ratio))
        
        noise = torch.rand(N, L, device=device)  # noise in [0, 1]
        
        # Sort noise for each sample
        ids_shuffle = torch.argsort(noise, dim=1)  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # Keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]

        # Generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=device)
        mask[:, :len_keep] = 0
        
        # Unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return mask, ids_restore, ids_keep

    def random_masking_plus_given_index(self, input_size, device, given_index, **kwargs):
        """
        Perform per-sample random masking by per-sample shuffling and also mask out the given index.
        input_size: [N, L, D], sequence
        given_index: [L_g], given index to be masked out
        device: torch.device
        mask: [N, L], binary mask
        ids_restore: [N, L], indices to restore the original order
        """
        N, L, D = input_size
        maskout_index_total = np.union1d(given_index, np.random.choice(L, int(L * self.mask_ratio), replace=False))
        len_keep = L - len(maskout_index_total)
        mask = torch.zeros([N, L], device=device)
        mask[:, maskout_index_total] = 1
        ids_shuffle = torch.argsort(mask, dim=1)
        ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :len_keep]
        return mask, ids_restore, ids_keep


class PixelShuffle3D(nn.Module):
    def __init__(self, image_size, patch_size, in_dim, out_dim, layer_type="linear", scale_factor=0.5):
        super(PixelShuffle3D, self).__init__()
        # Initialize the parameters for the PixelShuffle3D class
        self.image_size = image_size
        self.patch_size = patch_size
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.layer_type = layer_type
        self.patches_ori_tuple = [img // pch for img, pch in zip(image_size, patch_size)]
        self.num_patches_ori = np.prod(self.patches_ori_tuple)
        self.scale_factor = scale_factor
        self.num_patches_downsample = int(self.num_patches_ori * (scale_factor ** 3))
        self.shuffle_dim = int(in_dim / (scale_factor ** 3))
        # Linear or MLP projection for the input tokens
        if layer_type == "linear":
            self.proj = nn.Linear(self.shuffle_dim, self.out_dim)
        elif layer_type == "mlp":
            self.proj = nn.Sequential(
                nn.Linear(self.shuffle_dim, self.out_dim),
                nn.GELU(),
                nn.Linear(self.out_dim, self.out_dim)
            )
            
    def pixel_shuffle_3d(self, x):
        # Assuming the input x has shape (N, W, H, Z, C)
        x = x.contiguous()
        n, w, h, z, c = x.size()
        # Calculate new dimensions based on the scale factor
        new_w = int(w * self.scale_factor)
        new_h = int(h * self.scale_factor)
        new_z = int(z * self.scale_factor)
        new_c = int(c / (self.scale_factor ** 3))
        # N, W, H, Z, C --> N, W * scale, H * scale, Z * scale, C // (scale ** 3)
        x = x.view(n, new_w, new_h, new_z, new_c)
        # N, W * scale, H * scale, Z * scale, C // (scale ** 3) --> N, H * scale, W * scale, Z * scale, C // (scale ** 3)
        x = x.permute(0, 2, 1, 3, 4).contiguous()
        return x
    
    def forward(self, x):
        x = x.reshape(-1, *self.patches_ori_tuple, self.in_dim)
        x = self.pixel_shuffle_3d(x)
        x = x.permute(0, 4, 1, 2, 3)
        x = x.reshape(-1, self.num_patches_downsample, self.shuffle_dim)
        x = self.proj(x)
        return x
    
    
def pixel_shuffle3d(input, upscale_factor):
    batch_size, channels, time, height, width = input.size()
    channels //= upscale_factor ** 3
    output_time = time * upscale_factor
    output_height = height * upscale_factor
    output_width = width * upscale_factor
    
    input = input.view(batch_size, channels, upscale_factor, upscale_factor, upscale_factor, 
                       time, height, width)
    output = input.permute(0, 1, 5, 2, 6, 3, 7, 4).contiguous()
    output = output.view(batch_size, channels, output_time, output_height, output_width)
    
    return output


def pixel_unshuffle3d(input, downscale_factor):
    batch_size, channels, time, height, width = input.size()
    output_channels = channels * (downscale_factor ** 3)
    output_time = time // downscale_factor
    output_height = height // downscale_factor
    output_width = width // downscale_factor
    
    input = input.view(batch_size, channels, output_time, downscale_factor, 
                       output_height, downscale_factor, output_width, downscale_factor)
    output = input.permute(0, 1, 3, 5, 7, 2, 4, 6).contiguous()
    output = output.view(batch_size, output_channels, output_time, output_height, output_width)
    
    return output